// module uls.cpp
// Copyright Dr.E.Huckert 04-2008
// Receives UDP logging messages on standard port 9514
// The port can be changed via item "[general ]rxport" in file uls.ini 
// Sends (own) UDP logging messages via port 9515
// Do not use the same port for logging ans for receive!
//
// A logging server (=receiver) for UDP based logging
// Compiled with Digital Mars and wxWidgets V2.8.7 - OK 
// (see Makefile uls.mak)
// Needs source modules inifilewx.cpp and hu_trace.cpp

// versions
// 16.08.2014 adapted for Ubuntu/Linux, see makefile ulsu.mak
//            used with wxWidgets V3.0.1, see makefile uls.mak
//            re-tested under Windows with wxWidgets 2.8.7

#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>

#include "wx/wx.h"
#include "wx/string.h"
#include "wx/cursor.h"
#include "wx/list.h"
#include "wx/hashmap.h"

#include "hu_trace.h"
#include "inifilewx.h"

#ifdef WIN32
#include <winsock.h>
#include <io.h>
WSADATA WSAdata;
#else
#define wsprintf sprintf
#endif

#include "uls.xpm"               // an icon file in xpm format

#define WIDTH     500
#define HEIGHT    600
#define XPOS      (WIDTH/5)
#define YPOS      (HEIGHT/5)
#define LEDIT      80
#define LSTATIC   100
#define LBUTTON    60
#define CTLHEIGTH  26

// we receive UDP messages on this port:
int udpRxPort        = 9514;
wxString versionInfo =  "uls V1.3 by EH 08-2014";

#ifdef LOGGG
// we log (send )on port 9515
HU_TRACE *pTrace = new HU_TRACE("localhost", 9515);
#else
// default:no own logging
HU_TRACE *pTrace = NULL;
#endif
// global configuration
IniFileWx *pIni           = NULL;
static int continueThread = 1;
static int globalErrors   = 0;
char showLevel            = 0;
sock_t globalSocket       = -1;
wxListBox  *pGlobalList   = NULL;
int        writeToFile    = 0;

// ------------------------------------------------------------
class ulsThread : public wxThread
{
  public:
  ulsThread();
  void * Entry();
};

// ----------------------------------------------------------
// sleep a number of milliseconds
int milli_sleep(int millis)
{
int ret;
  ret = 0;
#ifdef WIN32
  ::Sleep(millis);
#else
  ret = usleep((long)millis * 1000L);
#endif
  return ret;
}   // end milli_sleep()

// --------------------- socket handling and WSA ------------------
int create_socket(int af, 
                  int type, 
                  int protocol ) 
{
    sock_t sock;
    const int y = 1;
    sock = ::socket(af, type, protocol);
    if (sock < 0)
    {
      return -1;
    }
    ::setsockopt( sock, SOL_SOCKET,
                SO_REUSEADDR, (char *)(&y), sizeof(int));
    return sock;
}

// ----------------------------------------------------------------
void close_socket( sock_t *sock )
{
  #ifdef WIN32
  closesocket(*sock);
  #else
  #endif
}
// ----------------------------------------------------------------
// Receive a portion of data via UDP
int UDP_recv(sock_t *sock, char *data, size_t size)
{
   struct sockaddr_in addr_recvfrom;
   int flags = 0;
   #ifdef WIN32
   int len;
   #else
   socklen_t len;
   #endif
   int n;
    //
   len = sizeof (addr_recvfrom);
   // note: flags=0 means: this is a blocking call - it will return only if
   //           data are available
   n   = ::recvfrom(*sock, data, size, flags,
                    (struct sockaddr *)&addr_recvfrom, 
                    &len);
   #ifdef WIN32
   if (n == SOCKET_ERROR)
   #else
   if (n < 0)
   #endif
   {
     if (::pTrace != NULL)
       ::pTrace->trace("UDP receive error");
     return -1;
   }
   return n;
}   // end UDP_recv()

// ----------------------------------------------------------------
int bind_socket(sock_t sock, 
                long   lAddress,
                unsigned short port)
{
  struct sockaddr_in server;
  struct hostent *h;
  ::memset(&server, 0, sizeof (server));
  server.sin_family      = AF_INET;
  server.sin_addr.s_addr = htonl(lAddress);  // nicht ::htonl
  server.sin_port        = htons(port);      // nicht ::htons
  if (::bind(sock, (struct sockaddr*)&server, sizeof(server)) < 0)
     return -1;
  return 0;
}
// ----------------------------------------------------------------
void cleanup(void)
{
   /* Cleanup Winsock */
   #ifdef WIN32
   WSACleanup();
   #else
   #endif
}

// ----------------------------------------------------------------
// extract the module name from a message
// typical sample for msg: <n>HH:MM:SS.mmm modulename ...............
wxString getModuleName(const char *msg)
{
  wxString modName = "";
  int n, noSpaces;
  noSpaces = 0;
  for (n=0; msg[n] != 0; n++)
  {
    if (msg[n] == ' ') 
    {
      noSpaces++;
      if (noSpaces > 1) break;
      continue;
    }
    if (noSpaces == 1)
    {
      modName.Append(msg[n]);
    }
  }
  return modName;
}   // end getModuleName()

// -------------------------------------------------------------------
// Eliminates unwanted characters from the buffer
void cleanBuffer(char *buf)
{
int n;
 if (buf == NULL)
   return;
 for (n=0; buf[n] != 0; n++)
 {
   if (buf[n] == '\n') buf[n] = 0;
   if (buf[n] == '\r') buf[n] = 0;
 }
}   // end clean_buffer()

// ------------------------------------------------------------------
// constructor
ulsThread::ulsThread()
{
}

// -------------------------------------------------------------------
// Thread procedure 
// Receives data from an UDP socket
// Shows the received data in a list box
// Assumes that the socket is open
void * ulsThread::Entry()
{
  static unsigned long ret = 0;
  int showIt;
  char puffer[1024];
  int level;
  wxString aux, line;
  FILE *ofd = NULL;
  unsigned warnCount = 0;
  //
  if (::pTrace != NULL)
    ::pTrace->trace("ulsThread::Entry entry");
  //
  //Note: continueThread may be set to 0 externally to stop the thread
  // No sleep() or a similar call is necessary to release the control to other threads
  // as UDP_recv() does this!
  while (::continueThread)
  {
    if (this->TestDestroy())
      // we are about to be killed - leave this loop immediately
      break;
    memset (puffer, 0, sizeof(puffer));
    // this is a blocking call - so ::continueThread may be useless!
    ret = UDP_recv(&::globalSocket, puffer, sizeof(puffer));
    if (::pTrace != NULL)
       ::pTrace->trace("UDP_recv() ret=", ret);
    if (ret < 0)
    {
      // ulsThread::Entry error
      if (::pTrace != NULL)
        ::pTrace->trace("ulsThread::Entry error");
      continue;   
    }
    cleanBuffer(puffer);
    showIt = 1;
    if (::showLevel != 0)
    {
      // check the message level
      // message syntax sample: "<5>date id text" where "<5>" is the message level
      if (puffer[0] == '<')
        level = puffer[1];
      else level = '9';
      if (level > showLevel)
        // message level is > than defined level
        showIt = 0;
    }
    if (showIt)
    {
      line = puffer;
      ::pGlobalList->InsertItems(1, &line, 0);
    }
    if (::writeToFile)
    {
      if (ofd == NULL)
      {
        // open the log output file
        #ifdef WIN32
        ofd = ::fopen(".\\uls.log", "w");
        #else
        ofd = ::fopen("./uls.log", "w");
        #endif
      }
      if (ofd == NULL)
      {
        if (warnCount++ == 0)
          ::wxMessageBox("Warning: cannot open log output file");
      }
      else
      {
        // write to the log file 
        ::fprintf(ofd, "%s\n", puffer);
        ::fflush(ofd);
      }
    }   // end if (::writeToFile)
  }   // end while
  //
  if (ofd != NULL)
    ::fclose(ofd);
  //if(pTrace != NULL) pTrace->trace("udp_receiver() UDP Thread terminated");
  #ifdef WIN32
  ExitThread(ret);
  #else
  #endif
  return (void *)&ret;
}   // end ulsThread::Entry()

// -----------------------------------------------------------
// Define a new application type
class ulsApp : public wxApp
{
public:
  virtual bool OnInit();
  void OnAssertFailure(const wxChar * file,
                       int  	line,
                       const wxChar *  	func,
                       const wxChar *  	cond,
                       const wxChar *  	msg);
  //bool ProcessEvent(wxEvent & event);
};

// -------------------------------------------------
// This is the windows where we draw on
class ulsWindow : public wxScrolledWindow
{
  public:
    ulsWindow(wxPanel *, long, wxPoint, wxSize);
    void draw();
    int getDrawStatus() { return this->isDrawn; };
  private:
    wxWindowDC *pdc;
    int isDrawn;
};   // end class ulsWindow

// ------------------------------------------------------------
class ulsFrame : public wxFrame
{
public:
  ulsFrame(int xp, int yp, int xl, int yl);
  ~ulsFrame();

  void  OnQuit(wxCommandEvent& event);
  void  OnAbout(wxCommandEvent& event);
  void  OnClear(wxCommandEvent& event);
  void  OnSetFocus(wxFocusEvent& event);
  void  OnActivate(wxActivateEvent& event);
  void  OnMove(wxMoveEvent & event);
  void  OnSize(wxSizeEvent & event);
  void  OnUseFile(wxCommandEvent &event);
private:
  ulsThread         *serverThread;
  wxListBox         *pListbox;
  wxPanel           *panel;
  wxMenu            *m_menuFile;
  wxMenuBar         *m_menuBar;
  wxColour          bgColour;
  wxColour          colour1;
  wxButton          *pButton1;
  wxCursor          stdCursor;
  wxTextCtrl        *portCtrl;
  sock_t            sock;
  int               port;
  //
  void              makePanel();
  wxStaticText *makeStaticText(const char *pLabel, 
                                int startx,
                                int starty);
  wxTextCtrl *makeEditField(const char *label,
                            int startx,
                            int starty,
                            long style);

  DECLARE_EVENT_TABLE()
};

enum
{
  BUTTON_CLEAR = 11000,
  SERVER_QUIT  = wxID_EXIT,
  USE_FILE     = 11001,
  SERVER_ABOUT = wxID_ABOUT
};

BEGIN_EVENT_TABLE(ulsFrame, wxFrame)
  EVT_MENU(SERVER_QUIT,   ulsFrame::OnQuit)
  EVT_MENU(USE_FILE,  ulsFrame::OnUseFile)
  EVT_MENU(SERVER_ABOUT,  ulsFrame::OnAbout)
  EVT_SET_FOCUS(ulsFrame::OnSetFocus)
  EVT_ACTIVATE(ulsFrame::OnActivate)
  //EVT_CLOSE(ulsFrame::OnQuit)              // TODO: wieder rein!
  EVT_MOVE(ulsFrame::OnMove)
  EVT_BUTTON(BUTTON_CLEAR, ulsFrame::OnClear)
  //EVT_SIZE(ulsFrame::OnSize)  // nicht aktivieren: Fehler!!!
END_EVENT_TABLE()

IMPLEMENT_APP(ulsApp)

// ---------------------------------------------------------------------
// Constructor
ulsWindow::ulsWindow(wxPanel *parent,   // ulsFrame *parent, 
                         long style, 
                         wxPoint position, 
                         wxSize size) 
    : wxScrolledWindow((wxFrame *)parent, -1, position, size, style)
{
  this->pdc     = NULL;
  this->isDrawn = 0;
  wxColour col("YELLOW");
  this->SetBackgroundColour(col);
  this->Show();
}   // end class ulsWindow

// ---------------------------------------------------------------------
void ulsWindow::draw()
{
  //if(pTrace != NULL) pTrace->trace("ulsWindow::draw()");
}


// ----------------------------------------------------------------
// noetig wegen obskurer CURSOR Probleme unter wxWidgets V3.0.1
void ulsApp::OnAssertFailure(const wxChar *  	file,
                            int  	line,
                            const wxChar *  	func,
                            const wxChar *  	cond,
                            const wxChar *  	msg) 	
{
  if (::pTrace != NULL) 
  {
    ::pTrace->trace("MyFrame::OnAssertFailure() file=", file);
    ::pTrace->trace("MyFrame::OnAssertFailure() line=", line);
    ::pTrace->trace("MyFrame::OnAssertFailure() func=", func);
  }
}   // end ulsApp::OnAssertFailure()

// ------------------------------------------------------------------
// Is invoked when the application starts
bool ulsApp::OnInit()
{
  long cRet = 0;
  wxString s;
  if (::pTrace != NULL)
    ::pTrace->trace("ulsApp::OnInit() 0");
  //
  // read the ini file uls.ini
  ::pIni = new IniFileWx();
  ::pIni->setLogging(::pTrace);
  #ifdef WIN32
  cRet = ::pIni->load(".\\uls.ini");
  #else
  cRet = ::pIni->load("./uls.ini");
  #endif
  if(pTrace != NULL)
    pTrace->trace("ulsApp::OnInit() IniFileWx::load() ret=", cRet);
  s = ::pIni->get("general", "rxport");
  if (s.length() > 0)
  {
    ::udpRxPort = ::atoi((const char *)(s.c_str()));
    if(pTrace != NULL)
      pTrace->trace("ulsApp::OnInit() udpRxPort=", ::udpRxPort);
  }
  //
  // Create the main application window
  ulsFrame *pFrame = new ulsFrame(XPOS, YPOS, WIDTH, HEIGHT);
  //if(pTrace != NULL) pTrace->trace("ulsApp::OnInit()1");
  pFrame->Show(true);
  //if(pTrace != NULL) pTrace->trace("ulsApp::OnInit() 2");
  SetTopWindow(pFrame); 
  //if(pTrace != NULL) pTrace->trace("ulsApp::OnInit() 3");
  // Success
  //if(pTrace != NULL) pTrace->trace("ulsApp::OnInit() 4");
  return true;
}   // end ulsApp::OnInit()

// -------------------------------------------------------------------
// frame constructor
ulsFrame::ulsFrame(int xp, int yp, int xl, int yl) 
    : wxFrame((wxFrame *)NULL, 
              wxID_ANY,
              versionInfo,
              wxPoint(xp, yp), 
              wxSize(xl, yl))
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 0");
  this->serverThread    = NULL;
  this->stdCursor       = 0;
  this->pListbox        = NULL;
  this->port            = ::udpRxPort;
  
  // Give the frame an icon
  //SetIcon(wxICON(uls));
  
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 1");
  this->bgColour = wxColour(200, 200, 200);
  this->colour1  = wxColour(64, 255, 64);

  // Make menus
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 2");
  m_menuFile = new wxMenu();
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 3");
  m_menuFile->Append(USE_FILE, 
                     _("log to &File\tCTRL-F"), 
                     _("Log to file"));
  m_menuFile->Append(SERVER_ABOUT, 
                     _("&About...\tCtrl-A"), 
                     _("Show about dialog"));
  m_menuFile->AppendSeparator();
  m_menuFile->Append(SERVER_QUIT, 
                    _("E&xit\tAlt-X"), 
                    _("Quit"));

  // Append menus to the menu bar
  m_menuBar = new wxMenuBar();
  m_menuBar->Append(m_menuFile, _("&File"));
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 4");
  SetMenuBar(m_menuBar);
  //
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 5");
  this->makePanel();
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 6");
  //
  #ifdef WIN32
  if (WSAStartup(0x0101, &(WSAdata)) != 0)
  {
    ::wxMessageBox("Error: cannot open UDP socket");
    exit(-1);
  }
  #endif
  // create an UDP socket
  this->sock = create_socket(AF_INET, SOCK_DGRAM, 0);
  if (this->sock < 0)
  {
    //cout << "Error create_socket()" << endl;
    globalErrors++;
  }
  ::globalSocket = this->sock;
  int ret = bind_socket(this->sock, INADDR_ANY, this->port);
  if (ret < 0)
  {
    //cout << "Error bind_socket()" << endl;
    globalErrors++;
  }
  //
  if (globalErrors == 0)
  {
    // create the ulsThread thread and run it
    this->serverThread = new ulsThread();
    this->serverThread->Create();
    if (! (this->serverThread)->IsRunning())
    {
      wxThreadError err;
      err = (this->serverThread)->Run();
      if (err != wxTHREAD_NO_ERROR)
      {
        wxMessageBox("Fehler", _("cannot start UDP RX thread"),
                 wxOK | wxICON_INFORMATION, this);
      }
    }
  }
  else
  {
    if (::pTrace != NULL)
      ::pTrace->trace("Error: cannot open UDP socket");
    ::wxMessageBox("Error: cannot open UDP socket");
    ::exit(-2);
  }
  //if(pTrace != NULL) pTrace->trace("ulsFrame::Constructor 7");
}   // end ulsFrame::ulsFrame()

// ----------------------------------------------------------------
// Destructor
ulsFrame::~ulsFrame()
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::destructor");
  //if(pTrace != NULL) delete pTrace;
  pTrace = NULL;
}

// ------------------------------------------------------------------
// create a static text control, i.e. a label
wxStaticText *ulsFrame::makeStaticText(const char *pLabel, 
                                      int startx,
                                      int starty)
{
  wxStaticText *pTextCtl;
  pTextCtl = new wxStaticText(this->panel, -1, pLabel,
                              wxPoint(startx, starty),
                              wxSize(LSTATIC, CTLHEIGTH));
  pTextCtl->SetBackgroundColour(this->bgColour);
  pTextCtl->Refresh();
  return pTextCtl;
}   // end ulsFrame::makeStaticText()

// -------------------------------------------------------------------
// Create a label and an edit field
// Return a pointer to the edit control
wxTextCtrl * ulsFrame::makeEditField(const char *label,
                                     int startx,
                                     int starty,
                                     long style)
{
  wxStaticText *pTextCtl;
  wxTextCtrl   *pEditCtl;
  pTextCtl = this->makeStaticText(label, startx, starty);
  pEditCtl = new wxTextCtrl(this->panel, 
                            -1, 
                            "", 
                            wxPoint(startx + LSTATIC + 5, starty),
                            wxSize(LEDIT, CTLHEIGTH),
                            style);
  //pEditCtl->SetEditable(true);
  return pEditCtl;
}   // end ulsFrame::makeEditField()

// ----------------------------------------------------------------
// create a panel i.e. the set of visible controls in the 
// main application window
void ulsFrame::makePanel()
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::makePanel()");
  char aux[128];
  this->panel  = new wxPanel(this);
  int xp = 5;
  int yp = 5 ;
  int xl;
  int yl; 
  this->GetClientSize(&xl, &yl); 
  //
  // an edit field that shows the UDP receive port
  this->portCtrl = this->makeEditField("  UDP RX Port:", 
                                       xp , yp,
                                       wxTE_READONLY | wxTE_CENTRE);
  wsprintf(aux, "%d", ::udpRxPort);
  this->portCtrl->SetValue(aux);
  // 
  // this button clears the actual contents of the listbox
  this->pButton1 = new wxButton(this->panel, 
                                BUTTON_CLEAR,
                                "Clear",
                                wxPoint(xl - LBUTTON - 10, yp),
                                wxSize(LBUTTON, CTLHEIGTH));
  pButton1->SetBackgroundColour(this->colour1);
  //
  // create a listbox where to show the UDP messages
  yp = yp + CTLHEIGTH + 5; 
  this->pListbox = new wxListBox(this->panel, 
                                  wxSUNKEN_BORDER, 
                                  wxPoint(xp, yp),
                                  wxSize(xl - 10, yl - yp - 5));
  ::pGlobalList = this->pListbox;
}   // end ulsFrame::makePanel()

// ----------------------------------------------------------------
void ulsFrame::OnSetFocus(wxFocusEvent& event)
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnSetFocus()");
  //this->pw->draw(); 
}

// ----------------------------------------------------------------
// Toggles the flag writeToFile
// Changes the frame title to indicate that a file is written
void ulsFrame::OnUseFile(wxCommandEvent& event)
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnUseFile()");
  if (! ::writeToFile) ::writeToFile = 1;
  else                 ::writeToFile = 0;
  if (::writeToFile)
    this->SetTitle(versionInfo + "   " +"file: ./uls.log");
  else
     this->SetTitle(versionInfo);
}

// ----------------------------------------------------------------
void ulsFrame::OnActivate(wxActivateEvent& event)
{
  //~ if (this->pw != NULL) 
    //~ this->pw->draw(); 
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnActivate() leaving");
}

// ----------------------------------------------------------------
void ulsFrame::OnMove(wxMoveEvent& event)
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnMove()");
  // get the actual position of the top left corner
  wxPoint pos = event.GetPosition();
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnMove() x=", pos.x);
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnMove() y=", pos.y);
  int xl, yl;
  // get the size of the window
  this->GetVirtualSize(&xl, &yl);
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnMove() size x=", xl);
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnMove() size y=", yl);
  //
  // check if the window is completely visible - if not we have to redraw the window!
  bool exp = this->IsExposed(pos.x, pos.y);
  //
  // redraw the inner window if it is not exposed
  if (! exp)
  {
    //~ if (this->pw != NULL)
    //~ {
      //~ if (this->pw->getDrawStatus()) 
        //~ this->pw->draw();
    //~ }
  }
}   // ulsFrame::OnMove()

// ----------------------------------------------------------------
// Vorsicht: wenn aktiviert anderes (problematisches) Verhalten! 
void ulsFrame::OnSize(wxSizeEvent& event)
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnSize()");
  //wxSize size = event.GetSize();
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnSize() width=", size.GetWidth());
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnSize() height=", size.GetHeight());
}

// ----------------------------------------------------------------
// The user has pressed the "Clear" button
void ulsFrame::OnClear(wxCommandEvent& event)
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnClear()");
  this->stdCursor = this->GetCursor();
  //
  // clear the list box
  this->pListbox->Set(0, NULL);
  this->pListbox->Update();    
}   // ulsFrame::OnClear()

// ----------------------------------------------------------------
void ulsFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnQuit()");
  // true is to force the frame to close
  //Close(true); // nicht aufrufen: gibt Schleife!!!!
  //
  // terminate the UDP receiver thread
  ::continueThread = 0;
  // wait some milliseconds so that the UDP receiver thread
  // can terminate
  ::milli_sleep(500);
  //
  if (this->serverThread != NULL)
  {
    /*
           // this doesn't work as we have a blocking call in the thread...
            if ((this->serverThread)->IsRunning())
            {
              this->serverThread->Delete();
            }
          */
    this->serverThread = NULL;
  }
  this->Destroy();
}   // end ulsFrame::OnQuit()

// ----------------------------------------------------------------
void ulsFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
  //if(pTrace != NULL) pTrace->trace("ulsFrame::OnAbout()");
  wxString s;
  s = "UDP based logging server\n";
  s = s  + "Copyright by Dr.E.Huckert 11-2007";
  wxMessageBox(s, _("About uls"),
               wxOK | wxICON_INFORMATION, this);
}   // end ulsFrame::OnAbout()
